/*

DESCRIPTION:    Control firmware for prototype 250W SEPIC coupled inductor 
                PS for use in a 48V lead-acid / LiFePO4 PV-based charger.
                Purpose is to validate design and proceed to interleave two to
                obtain 500W with low input & output ripple.

                Specifications:
                    + PV input voltage: 10 - 80 volts
                    + PV input current: nominal 4.3A
                    + Charging voltage: 56V
                    + Charging current: 4.5A (max)
                    + Maximum continuous power: 250W
                    + Control mode: current
                    + Bias supply: 12V
                    + Operating temperature: -20 - +50C (not validated)

                External and internal slope compensation ramps are included to 
                allow for experimentation and to validate what scheme is 
                necessary for the next design iteration. The internal ramp uses 
                the PRG which has a minimum slope of 0.2V/uS and allows for 
                software selectable increments up to 2.5V/uS. It applies a 
                negative slope to the voltage (outer loop) error amp. It 
                requires no external components and can be easily configured in
                firmware.

                The external slope compensation is derived from a linear ramp 
                that applies a positive slope to the current sense signal (the
                voltage EA signal is not modified). A voltage divider determines
                the slope compensation value, but it allows for smaller values
                than the internal scheme. It requires several additional 
                components and slope changes require component swaps.

                The firmware allows selection of either scheme via console 
                commands; the system can also be run open-loop (both 
                disconnected). Engagement of each is done by connecting the
                respective comparator output to the COG as a falling event (FE).

                An RS232 interface shared with the ICSP circuit provides an 
                optional console interface with a command set to allow 
                experimentation without firmware modification.  Many parameters 
                are stored in Non-Volatile Memory (NVM).

                RS232 settings: 115.2kbps, 8 data bits, 1 stop, no parity. This 
                console port is accessed over the micro-USB connector on  the 
                PCB and is optimized for use with a GBIB-style control program.

                A status LED on the PCB indicates if the charger is running 
                (Green) or in a stopped or fault state (RED).

SYNTAX:         n/a

KNOWN BUGS:     n/a

COPYRIGHT:      (c) 2020, Brian Cornell, all rights reserved.

DEPENDENCIES:   see include file listing.

 TARGET MCU:    PIC16F1769

HW REVISION:    200713

HW RELEASE:     prototype

VERSION:        0.1

MODIFICATION HISTORY:

 
    3 Aug 20, BWC, adapted baseline from mpptproto project.
    

*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * COMPILER & CLIB INCLUDES.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FIRMWARE INCLUDES:
 * + _xtal_freq.h:  sets the constant _XTAL_FREQ to the system oscillator
 *                  frequency (Fosc).  Used by the __delayxxx macros.
 * + binary.h:      Boolean constants shared across firmware projects.
 * + build.h:       managed by the script build.sh and called by the make file.
 *                  Maintains the compile build number.
 * + config.h:      PIC config registers
 * + dspstatus.h:   system status & display definitions.  Shared with project
 *                  rs232-44780 (used as optional status display).
 * + mpptproto.h:   main include file for this project.
 * + usart.h:       usart module function declarations, constants, etc.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include <binary.h>
#include "config.h"
#ifndef _XTAL_FREQ
#include "_xtal_freq.h"
#endif
#include "usart.h"
#include "sepicproto.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * MODULE DATA.
 * + Opdata is a module-level object that maintains dynamic operating data (e.g.
 *   voltages, currents, etc.).
 * + Opdata_ss is a snapshot of Opdata.
 * + flag_ss is TRUE if a snapshot of system data exists in object Opdata_ss.
 * + flag_nvm is TRUE when NVM must be updated due to console input.
 * + RDword enables word access to the ramData object for 16-bit NVM values.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
sysdat Opdata,
       Opdata_ss;
__bit flag_ss;
__bit flag_nvm;
uint16_t *RDword;


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Non-volatile parameter storage & management.  nvData resides in program flash
 * at the top of the high-endurance section (to make the lower portion 
 * contiguous & available for code).  ramData is the working copy in RAM.
 * 
 * Initialization of nvData & ramData MUST match: these are system defaults and 
 * enable recovery after NVM corruption or to restore via console commands.  
 * Refer to detailed explanation of NVM management at beginning of function 
 * Initialize.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const unsigned char nvData[FLASHROWSZ] __at(0x1FE0) = {S_DF_FLASH_0, S_DF_FLASH_1, \
S_DF_FLASH_2, S_DF_FLASH_3, S_DF_FLASH_4, S_DF_FLASH_5, S_DF_FLASH_6, \
S_DF_FLASH_7, S_DF_FLASH_8, S_DF_FLASH_9, S_DF_FLASH_10, S_DF_FLASH_11, \
S_DF_FLASH_12, S_DF_FLASH_13, S_DF_FLASH_14, S_DF_FLASH_15, S_DF_FLASH_16, \
S_DF_FLASH_17, S_DF_FLASH_18, S_DF_FLASH_19, S_DF_FLASH_20, S_DF_FLASH_21, \
S_DF_FLASH_22, S_DF_FLASH_23, S_DF_FLASH_24, S_DF_FLASH_25, S_DF_FLASH_26, \
S_DF_FLASH_27, S_DF_FLASH_28, S_DF_FLASH_29};
unsigned char ramData[FLASHROWSZ] = {S_DF_FLASH_0, S_DF_FLASH_1, \
S_DF_FLASH_2, S_DF_FLASH_3, S_DF_FLASH_4, S_DF_FLASH_5, S_DF_FLASH_6, \
S_DF_FLASH_7, S_DF_FLASH_8, S_DF_FLASH_9, S_DF_FLASH_10, S_DF_FLASH_11, \
S_DF_FLASH_12, S_DF_FLASH_13, S_DF_FLASH_14, S_DF_FLASH_15, S_DF_FLASH_16, \
S_DF_FLASH_17, S_DF_FLASH_18, S_DF_FLASH_19, S_DF_FLASH_20, S_DF_FLASH_21, \
S_DF_FLASH_22, S_DF_FLASH_23, S_DF_FLASH_24, S_DF_FLASH_25, S_DF_FLASH_26, \
S_DF_FLASH_27, S_DF_FLASH_28, S_DF_FLASH_29};



/*Function main.  Control program entry point (from root code) for Maximum Power 
 * Point Tracking (MPPT) 12V Photo-Voltaic (PV) charger prototype.  Manages 
 * boot, operation, fault handling, user interface, and shutdown.
 * 
 * Consists of two functional blocks:
 *  + Initialization
 *  + RUN loop
 * 
 * Initialization does just that:  initializes the MCU & operating environment 
 * and performs the necessary boot & diagnostic checks.  When done it always
 * turns control over to the RUN loop.
 * 
 * The RUN loop is an infinite loop that uses waterfall processing.  Several
 * discrete code blocks are ordered according to dependency and / or priority:
 *  + PV current sense calibration
 *  + Data collection
 *  + Fault monitoring
 *  + Safety shutdown
 *  + RUN monitoring & control
 *  + BOOT (autostart)
 *  + Console service
 * 
 * Each block is documented.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * FUNCTION DATA.
     * + cmdb points to the start of the console command buffer.
     * + cmdp points to the start of parameters (after the command) in the
     *   command buffer.
     * + cmd_ack points to the acknowledgement string to be sent in response to
     *   a command - either 'Ack' or 'Nck'.
     * + counter is a general purpose loop counter and scratch variable.  It 
     *   WILL NOT be used for multiple iterations thru the RUN loop.
     * + reg_status is assigned the value of the STATUS register at boot.  Refer
     *   to the 'MCU FAILURE RECORDING' block for more detail.
     * + reg_pcon is assigned the value of the PCON register at boot.  Refer
     *   to the 'MCU FAILURE RECORDING' block for more detail.
     * + cmdi is the numeric value of the string pointed to by object cmdp.
     *   Used by the CONSOLE SERVICE block.
     * + ss_ptr points to either the Opdata or Opdata_ss objects. Used by the
     *   CONSOLE SERVICE commands.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    char *cmdb,                         //Start of command buffer
         *cmdp;                         //Start of command buffer parameters
    const char *cmd_ack;                //Pointer to command acknowledge string
    uint8_t counter = 0,                //General use counter
            reg_status,                 //Copy of STATUS register
            reg_pcon;                   //Copy of PCON register
    unsigned int cmdi;                  //Console parameter as unsigned int
    sysdat  *ss_ptr;                    //Points to Opdata or Opdata_ss object


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * MCU FAILURE RECORDING.
     * 
     * If an MCU failure (e.g. stack overflow, watchdog, etc.) is detected at 
     * boot it's important to understand what caused the fault.  The STATUS &
     * PCON registers are saved at the start of the main function.  Once 
     * initialization is complete, but before the run loop, these are saved to 
     * object ramData for writing to NVM if the MCU status code is set.
     * 
     * They will remain in NVM until another MCU trap occurs, the NVM status
     * code is cleared, or defaults restored.  The console command for get NVM 
     * status will return the status code mnemonic along with the STATUS & PCON
     * registers, respectively, in hexadecimal.  Register values are only
     * recorded when an MCU status is set so these will be 0x00 for any other
     * code.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    reg_status = STATUS;
    reg_pcon = PCON;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZATION.
     * 
     * Function Initialize configures the MCU and peripherals.  Afterwards a 
     * delay provides settle time.  Local variables are initialized before
     * entering the RUN loop.
     * 
     * Initialize local variables:
     * + Save STATUS & PCON registers in ramData object for subsequent write to
     *   NVM if MCU error by the 'fault monitoring & recovery control' block
     *   within the run loop.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Initialize();
    __delay_ms(S_TM_INT);

    if (Opdata.sts == STS_MCU)
    {
        ramData[FLASH_AR_STATUS] = (unsigned char) reg_status;
        ramData[FLASH_AR_PCON] = (unsigned char) reg_pcon;
    }


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * RUN LOOP PRIORITIZATION & WATCHDOG MANAGEMENT.
     * 
     * Once entered, the run-time loop's only exit points are a reboot or power
     * cycle.  The loop consists of several discrete blocks responsible for 
     * data collection, monitoring, and control.  They are organized in order
     * of criticality and/or dependency of subsequent blocks.  Categorically the
     * order is:
     *      + PV current sense calibration
     *      + Data collection
     *      + Fault monitoring & recovery control (includes NVM status 
     *        recording)
     *      + Shut-down
     *      + Run monitoring & control (including MPPT algorithm)
     *      + Boot (autostart)
     *      + Console service
     * 
     * The watchdog timer is expected to be on in the run loop regardless of
     * state.  It should be set to a value that accounts for the longest loop
     * time with at least a 100% margin.  If it must be disabled in anticipation
     * of a long delay then it is the responsibility of that code block to 
     * re-enable.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    SWDTEN = TRUE;

    while (TRUE)
    {
        CLRWDT();                           //Clear watchdog timer

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * DATA COLLECTION.
         * 
         * This block measures system voltages, over-current count, and COG 
         * timing. Single samples are taken of the input & output voltages. No 
         * attempt is made to filter noise / ripple since this is a prototype 
         * and these values are not used for digital control.
         * 
         * A read-safe copy of the system ISR-maintained time is made to 
         * represent the approximate time that data collection was performed. 
         * Interrupts are disabled for data integrity.
         * 
         * Timer1 counts the over-current events via logic pulses delivered to
         * a pin via comparator two's (C2) output. This eliminates the overhead
         * of interrupting the MCU which can cause resets when successive, 
         * sustained over-current events are encountered. Only when timer1
         * overflows is an interrupt generated to set a status bit. Since the
         * timer is 16 bits it is disabled when fetching the count to prevent 
         * corruption which may cause the loss of an event.
         * 
         * The COG timing (duty cycle) is measured using timer5's gating feature
         * to control counting. The gate pin is connected to COG1-A output. To
         * start timing the timer is cleared and the go bit set. Timing begins
         * on the next rising edge of the COG1-A pin and stops when it falls.
         * Timing is limited to S_TM_TMR_5_RUN time and is considered invalid
         * if exceeded. This allows the block to execute when the system is 
         * idle (COG1 off) without accommodation.
         * 
         * The raw timer value is stored in Opdata.cogc. The composite time is
         * stored in a single byte with the high nibble representing the integer
         * and the low the decimal. Hence, resolution is limited to 10th's uS.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReadVoltage(PIN_IA_IVOLT, S_RT_IN, &Opdata.in);     //Input voltage
        ReadVoltage(PIN_IA_OVOLT, S_RT_OUT, &Opdata.out);   //Output voltage

        //Record time data collection was done.
        di();
        Opdata.s_msc = Opdata.msc;
        Opdata.s_sc = Opdata.sc;
        ei();

        //Get over-current count
        T1ON = FALSE;                           //Disable to prevent corruption
        Opdata.ovc = TMR1;
        T1ON = TRUE;                            //Re-enable timer1

        //System (COG1) duty cycle
        {
            //Clear timer: high byte is not used
            TMR5 = 0;

            T5GGO_nDONE = TRUE;                 //Start capture
            __delay_us(S_TM_TMR_5_RUN);         //Fixed wait time for completion
            if (T5GGO_nDONE)                    //Problem or COG is off
            {
                T5GGO_nDONE = FALSE;            //Stop timing
                TMR5L = (S_QY_TMR_5_MAX + 1);   //Set high value & let remaining logic clear
            }

            if (TMR5L > S_QY_TMR_5_MAX)         //Invalid DC - clear values
            {
                Opdata.cogc = 0;
                Opdata.cogti = 0;
                Opdata.cogtd = 0;
            }
            else                                //Calculate valid DC
            {
                Opdata.cogc = TMR5;
                Opdata.cogti = (TMR5L / S_QY_TMR_5_1US);
                Opdata.cogtd = (((TMR5L % S_QY_TMR_5_1US) * 10) / \
                        S_QY_TMR_5_1US);
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * SHUT-DOWN.
         * 
         * This is a safety interlock to stop the charger if the COG state and
         * status do not agree. getSnapshot takes a snapshot of the system 
         * status that triggered the shutdown for console retrieval.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((cogStatus()) && (Opdata.sts != STS_RUN))
        {
            getSnapshot(TRUE);
            ChargerOFF();
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * NVM STATUS RECORDING.
         * 
         * Unrecoverable FAULT states are written to NVM and unit is halted. 
         * Make sure Vdd is sufficient and attempt write to NVM.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((Opdata.sts > STS_UND) && (ramData[FLASH_AR_STS] != Opdata.sts))
        {
            ReadVoltage(ADC_CH_DAC_3, S_RT_VDD, &Opdata.vdd);
            if (Opdata.vdd.volt > S_LM_VDD_NVM)
            {
                ramData[FLASH_AR_STS] = Opdata.sts;
                writeFlashRow((uint16_t) nvData, ramData);
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * RUN MONITORING & CONTROL.
         * 
         * Place holder.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * BOOT (AUTOSTART).
         * 
         * This block is responsible for completing the boot process. If 
         * autostart is set (FLASH_AR_AUTO) the charger is started; otherwise 
         * the STOP status code is set and the charger must be started via 
         * console commands.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Opdata.sts == STS_BOOT)
        {
            getSnapshot(TRUE);
            if (ramData[FLASH_AR_AUTO]) ChargerON();
            else Opdata.sts = STS_STOP;
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * CONSOLE SERVICE.
         * 
         * IMPORTANT:  do not place any code expected to execute on each RUN 
         * loop iteration beyond this point.  The console service handler should 
         * always be placed at bottom of run loop.
         * 
         * This block services the user console accessed via the RS232 micro-USB
         * service port.  The interface is styled after GPIB instruments.  All 
         * commands consist of a two-byte NL (new line) terminated string and 
         * are categorized as get, set, or action commands.  Refer to the 
         * definition of object CmdList for command strings, explanation, and 
         * values.
         * 
         * For efficiency the command line is parsed so that the command and
         * parameter are available as strings and integer (parameter only).  All
         * commands take no more than one parameter.
         * 
         * Some of the 'Get' commands support retrieval of system values from
         * the snapshot object Opdata_ss.  This object is of the same type as 
         * Opdata and represents the state of the system when an event occurred 
         * to be used for troubleshooting or experimentation.  The shut-down & 
         * boot blocks capture a system-generated snapshot.  The console also 
         * contains a command to capture a snapshot.  The system bit flag 
         * flag_ss governs the capture of a snapshot:  if set neither the system 
         * or console will overwrite the contents of the snapshot object.  This 
         * is done to preserve the contents for diagnostics.  The flag can only 
         * be cleared with a console command or reboot.  Note that a console 
         * initiated snapshot does not set the flag and can be over-written by a 
         * system event.
         * 
         * Snapshot values are retrieved by including a non-zero value after
         * the command. Doing so will return the data contained in the snapshot
         * object regardless of validity. The following commands can provide
         * snapshot values:
         *  + GC (Get over-current count)
         *  + GD (Get (dynamic) duty cycle %)
         *  + GO (Get output volts)
         *  + GV (Get input volts)
         * 
         * A value of zero or omitting the value will return 'real-time' data.
         * 
         * The firmware maintains non-precision system time in elapsed seconds
         * & 10's of mS. Seconds are maintained in a 16-bit integer so elapsed
         * time is limited to 65,535 seconds or ~ 10H. Time then overflows back
         * to zero. Most command responses will print this time preceeding the
         * command value in the format SSSSS.mS. mS are in ten's of mS so the
         * maximum value that should be displayed is 90 (900mS).
         * 
         * This time value is provided as a diagnostic reference. Snapshot 
         * values return with the approximate time that the snapshot was taken;
         * all others are 'real-time'.
         * 
         * For coding efficiency, most returned values will be numeric in the 
         * format 99.99. Expected boolean values will be 1.0=TRUE, 0.0=FALSE.
         * The decimal portion for non-boolean responses will be zero with two
         * exceptions:
         *  + Over-current event count will be displayed in the decimal position
         *    with a 1 in the integer position if the counter has overflowed.
         *  + Duty cycle is displayed uS with the decimal portion rounded to the
         *    ten's position.
         * 
         * All commands are evaluated in the following if/else block.  Commands 
         * that don't return a value respond with the string 'Ack' (success) or
         * 'Nck' (fail).  Character echo is off by default.
         * 
         * Individual commands that update NVM set the flag_nvm so that a single
         * call to writeFlashRow is made at the end of the console service 
         * block. The flag is cleared at the start of the block.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

        //Retrieve command buffer & parse
        if (usartStatRx() != USARTSTSRXRDY) continue;
        cmdb = (char *) usartGetRx();           //Get command buffer
        cmdp = cmdb;
        rmvSpaces(cmdp);                        //Command parameter as string
        cmdi = (unsigned int) atoi(cmdp);       //Parameter as unsigned int
        if (cmdi) ss_ptr = &Opdata_ss;          //Get snapshot data
        else ss_ptr = &Opdata;                  //Get real-time data
        cmd_ack = NULL;                         //Clear command result message
        flag_nvm = FALSE;                       //Clear NVM update flag

        //Get over-current count
        if (!(strncmp(cmdb, CmdList[CMDGETCNT], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) ss_ptr->ovc_of, (unsigned int) ss_ptr->ovc);
        //Get (dynamic) duty cycle
        else if (!(strncmp(cmdb, CmdList[CMDGETDCY], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) ss_ptr->cogti, (unsigned int) ss_ptr->cogtd);
        //Get max/set duty cycle
        else if (!(strcmp(cmdb, CmdList[CMDGETMDC])))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) ((ramData[FLASH_AR_DC] * 100) / S_TM_MOD_F),\
                    (unsigned int) 0);
        //Get set output volts
        else if (!(strcmp(cmdb, CmdList[CMDGETOVS])))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) DAC1REF, (unsigned int) 0);
        //Get snapshot of operating data
        else if (!(strncmp(cmdb, CmdList[CMDGETSNP], CMDSIZSTR)))
        {
            if (getSnapshot(FALSE)) cmd_ack = msgAck;
            else cmd_ack = msgNck;
        }
        //Get over-current limit
        else if (!(strcmp(cmdb, CmdList[CMDGETCUR])))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) DAC2REF, (unsigned int) 0);
        //Get loop setting
        else if (!(strcmp(cmdb, CmdList[CMDGETLPS])))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) ramData[FLASH_AR_LOOP], (unsigned int) 0);
        //Get ramp type (E | I)
        else if (!(strcmp(cmdb, CmdList[CMDGETRPS])))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) ramData[FLASH_AR_RAMP_E], (unsigned int) 0);
        //Get NVM status code
        else if (!(strcmp(cmdb, CmdList[CMDGETNVM])))
            printf("%s %X %X\n", StsList[ramData[FLASH_AR_STS]], \
                    (unsigned int) ramData[FLASH_AR_STATUS], \
                    (unsigned int) ramData[FLASH_AR_PCON]);
        //Get output voltage
        else if (!(strncmp(cmdb, CmdList[CMDGETOUT], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    ss_ptr->out.volt, ss_ptr->out.mv);
        //Get firmware release
        else if (!(strcmp(cmdb, CmdList[CMDGETREL])))
            printf(fmtPrg, VERMAJOR, VERMINOR, VERBUILD);
        //Get system status
        else if (!(strncmp(cmdb, CmdList[CMDGETSTS], CMDSIZSTR)))
            printf("%s\n", StsList[ss_ptr->sts]);
        //Get autostart setting
        else if (!(strcmp(cmdb, CmdList[CMDGETAUT])))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) ramData[FLASH_AR_AUTO], (unsigned int) 0);
        //Get internal ramp setting
        else if (!(strcmp(cmdb, CmdList[CMDGETIRS])))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    (unsigned int) PRG1CON2, (unsigned int) 0);
        //Get input voltage
        else if (!(strncmp(cmdb, CmdList[CMDGETINV], CMDSIZSTR)))
            printf(fmtGpib, ss_ptr->s_sc, (unsigned int) ss_ptr->s_msc, \
                    ss_ptr->in.volt, ss_ptr->in.mv);
        //Clear [current] counter
        else if (!(strcmp(cmdb, CmdList[CMDCLRCNT])))
        {
            //Disable when clearing to prevent corruption.
            T1ON = FALSE;
            TMR1 = 0;
            Opdata.ovc_of = 0;
            T1ON = TRUE;
            cmd_ack = msgAck;
        }
        //Clear NVM error
        else if (!(strcmp(cmdb, CmdList[CMDCLRNVM])))
        {
            ramData[FLASH_AR_STS] = STS_UND;
            ramData[FLASH_AR_STATUS] = (unsigned char) NULL;
            ramData[FLASH_AR_PCON] = (unsigned char) NULL;
            flag_nvm = TRUE;
            cmd_ack = msgAck;
        }
        //Clear snapshot flag
        else if (!(strcmp(cmdb, CmdList[CMDCLRSNP])))
        {
            flag_ss = FALSE;
            cmd_ack = msgAck;
        }
        //Restore NVM defaults
        else if (!(strcmp(cmdb, CmdList[CMDRSTDFL])))
        {
            writeFlashRow((uint16_t) nvData, NULL);
            cmd_ack = msgAck;
        }
        //Reset MCU
        else if (!(strcmp(cmdb, CmdList[CMDRSTMCU])))
        {
            //Reset system; first do clean stop if running
            if (cogStatus()) ChargerOFF();
            SWDTEN = FALSE;             //Disable so watchdog doesn't erroneously reboot
            PCON = S_LM_PCON_BT;        //Clear PCON bits to clear error trap on boot
            __delay_ms(S_TM_RST);
            RESET();
        }
        //Set autostart
        else if (!(strncmp(cmdb, CmdList[CMDSETAUT], CMDSIZSTR)))
        {
            if (cmdi) ramData[FLASH_AR_AUTO] = TRUE;
            else ramData[FLASH_AR_AUTO] = FALSE;
            flag_nvm = TRUE;
            cmd_ack = msgAck;
        }
        //Set per-cycle current limit
        else if (!(strncmp(cmdb, CmdList[CMDSETCUR], CMDSIZSTR)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Set the current limit by updating the DAC with user-supplied 
             * value.  The entered value is bounds checked.  When changed the 
             * updated value is stored in NVM.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmd_ack = msgAck;
            if ((cmdi < S_LM_CLLOW) || (cmdi > S_LM_CLHIG)) cmd_ack = msgNck;
            else if (cmdi != *(RDword + FLASH_AR_CLIM))
            {
                *(RDword + FLASH_AR_CLIM) = cmdi;
                flag_nvm = TRUE;
                DAC2REFH = ramData[FLASH_AR_CLIM_H];
                DAC2REFL = ramData[FLASH_AR_CLIM_L];
                DAC2LD = TRUE;
                while (DAC2LD);
            }
        }
        //Set duty cycle
        else if (!(strncmp(cmdb, CmdList[CMDSETDCY], CMDSIZSTR)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Convert user-entered duty cycle percentage to the PWM register 
             * value, bound check, and store in NVM if updated.  If running, the
             * DC is changed using chgDutyCycle to avoid a transient.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmd_ack = msgAck;
            cmdi = (((cmdi * CFG_PWM_5_PERIOD) / S_RT_PERCENT) + 1);
            if ((cmdi < S_DC_MIN) || (cmdi > S_DC_MAX)) cmd_ack = msgNck;
            else if (cmdi != ramData[FLASH_AR_DC])
            {
                //Update NVM
                ramData[FLASH_AR_DC] = (unsigned char) cmdi;
                flag_nvm = TRUE;
                //If operating move to new DC now
                if (!(COG1ASD0bits.ASE))
                    chgDutyCycle((uint16_t) PWM5DC, cmdi);
            }
        }
        //Set console echo on / off.  Anything other than 0 turns on.
        else if (!(strncmp(cmdb, CmdList[CMDSETECH], CMDSIZSTR)))
        {
            if (cmdi) usartSetEcho(TRUE);
            else usartSetEcho(FALSE);
            cmd_ack = msgAck;
        }
        //Set loop open / closed
        else if (!(strncmp(cmdb, CmdList[CMDSETLOP], CMDSIZSTR)))
        {
            if (cogStatus()) cmd_ack = msgNck;          //Must be off
            else
            {
                if (cmdi)                               //Closed
                {
                    ramData[FLASH_AR_LOOP] = TRUE;
                    setloopClosed();
                }
                else                                    //Open
                {
                    ramData[FLASH_AR_LOOP] = FALSE;
                    setloopOpen();
                }
                if (ramData[FLASH_AR_LOOP] != nvData[FLASH_AR_LOOP])
                    flag_nvm = TRUE;
                cmd_ack = msgAck;
            }
        }
        //Set ramp type (1=E | 0=I)
        else if (!(strncmp(cmdb, CmdList[CMDSETRPS], CMDSIZSTR)))
        {
            if (cogStatus()) cmd_ack = msgNck;          //Must be off
            else
            {
                //Update NVM setting...
                if (cmdi) ramData[FLASH_AR_RAMP_E] = TRUE;  //External
                else ramData[FLASH_AR_RAMP_E] = FALSE;      //Internal

                //But only apply to unit if set to closed loop
                if (ramData[FLASH_AR_LOOP])
                {
                    setloopClosed();
                }
                if (ramData[FLASH_AR_RAMP_E] != nvData[FLASH_AR_RAMP_E])
                    flag_nvm = TRUE;
                cmd_ack = msgAck;
            }
        }
        //Set charger on / off
        else if (!(strncmp(cmdb, CmdList[CMDSETCON], CMDSIZSTR)))
        {
            cmd_ack = msgAck;
            if (cmdi)
            {
                if (!(ChargerON())) cmd_ack = msgNck;
            }
            else ChargerOFF();
        }
        //Set internal ramp setting
        else if (!(strncmp(cmdb, CmdList[CMDSETIRS], CMDSIZSTR)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Set the ramp by updating the PRG with user-supplied value.  The 
             * entered value is bounds checked and the charger must be off. 
             * When changed the updated value is stored in NVM.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmd_ack = msgAck;
            if ((cmdi > S_LM_RHIG) || (cogStatus()))
                cmd_ack = msgNck;
            else if (cmdi != ramData[FLASH_AR_RAMP])
            {
                ramData[FLASH_AR_RAMP] = (unsigned char) cmdi;
                PRG1CON2 = ramData[FLASH_AR_RAMP];
                flag_nvm = TRUE;
            }
        }
        //Set output voltage
        else if (!(strncmp(cmdb, CmdList[CMDSETVOL], CMDSIZSTR)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Set the output voltage by updating the DAC with user-supplied 
             * value.  The entered value is bounds checked.  When changed the 
             * updated value is stored in NVM.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            cmd_ack = msgAck;
            if ((cmdi < S_LM_VLOW) || (cmdi > S_LM_VHIG)) cmd_ack = msgNck;
            else if (cmdi != *(RDword + FLASH_AR_VSET))
            {
                *(RDword + FLASH_AR_VSET) = cmdi;
                DAC1REFH = ramData[FLASH_AR_VSET_H];
                DAC1REFL = ramData[FLASH_AR_VSET_L];
                DAC1LD = TRUE;
                while (DAC1LD);
                flag_nvm = TRUE;
            }
        }

        if (flag_nvm) writeFlashRow((uint16_t) nvData, ramData);
        if (cmd_ack) printf(cmd_ack);   //Print command result
        usartSetRx(ENABLE);             //Look for new user input
    }
} //End function main


/* Function ChargerON.  Starts charger; if already started it has no effect.
 * No pre-start diagnostic voltage checks are done since this is expected to be
 * handled by the RUN loop and reflected in the value of object Opdata.sts.
 * 
 * Soft start is employed: the duty cycle is started at S_DC_MIN and gradually 
 * incremented to the DC stored in NVM. If open loop this is the DC the unit
 * will operate at, and if closed the control loop will regulate DC up to the
 * maximum set by NVM.
 * 
 * Returns:  TRUE if started; Opdata.sts set to RUN.
 *           FALSE if not; Opdata.sts contains error code.
 *
 * Parameters:  none
 *
*/
uint8_t ChargerON()
{
    uint8_t result;

    result = FALSE;

    if (COG1ASD0bits.ASE && ((Opdata.sts == STS_STOP) || \
            (Opdata.sts == STS_BOOT)))
    {
        TMR1 = 0;                               //Clear over-current count
        PRG1CON0bits.GO = TRUE;                 //Start slope compensation ramp
        PWM5DC = S_DC_MIN;                      //Soft start with min DC
        PWM5LDCONbits.LDA = TRUE;
        COG1ASD0bits.ASE = FALSE;               //Output on

        chgDutyCycle(S_DC_MIN, (uint16_t) ramData[FLASH_AR_DC]);

        Opdata.sts = STS_RUN;
        setledRUN();
        result = TRUE;
    }

    return(result);
} //End function ChargerON


/* Function ChargerOFF.  Stops charger by disabling the COG. The PRG is stopped 
 * to reduce bias current and switching noise when idle.
 * 
 * The status is set to STOP only if was set to RUN to preserve the condition
 * that initiated the STOP.
 * 
 * Returns:  none.  Opdata.sts is set to STOP provided it doesn't contain an 
 *           error code.
 *
 * Parameters:  none
 *
*/
void ChargerOFF()
{
    COG1ASD0bits.ASE = TRUE;
    PRG1CON0bits.GO = FALSE;                    //Stop slope compensation ramp
    if (Opdata.sts == STS_RUN) Opdata.sts = STS_STOP;
    setledFAULT();

    return;
} //End function ChargerOFF


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function getSnapshot. Copies the system object Obdata to Opdata_ss (snapshot
 * object) and optionally sets the snapshot flag. A snapshot will not be taken
 * if the system snapshot flag, flag_ss, is already set.
 * 
 * Only one snapshot can be taken and is not overwritten until flag_ss is 
 * cleared via console command.  Only system initiated events should set the 
 * flag. A console command to take a snapshot will fail if the system has 
 * already saved one (and set flag_ss TRUE).  Likewise, if the console does
 * take a snapshot and a subsequent system event occurs before it is read, the
 * system snapshot will overwrite the console copy.
 * 
 * Returns:  TRUE if snapshot taken; otherwise FALSE.
 *
 * Parameters:  set_flag=TRUE to set snapshot flag or FALSE.
 *
*/
__bit getSnapshot(uint8_t set_flag)
{
    if (!(flag_ss))
    {
        di();
        memcpy((void *) &Opdata_ss, (void *) &Opdata, sizeof(sysdat));
        ei();
        flag_ss = set_flag;

        return(TRUE);
    }

    return(FALSE);
}

/* Function ReadVoltage.  Reads the value from the analog channel, interprets as
 * a scaled voltage, and stores in the volts object pointed to by rv.  Three
 * Vref ranges can be selected depending on the required accuracy / resolution:
 *      + S_RT_ADC_MV_L (0-2V), 2mV/bit
 *      + S_RT_ADC_MV_H (0-4V), 4mV/bit
 *      + S_RT_ADC_MV_M (0-5V), 4.88mV/bit
 * 
 * The integer, decimal, and VMV components are stored in separate elements of 
 * the object pointed to by rv. All three are stored with the resolution 
 * defined by system constants. See details in the 'ADC configuration, 
 * acquisition, and value storage' section of this source' include file.
 * 
 * No error checking is performed and only a single sample is taken. The ADC
 * remains enabled on the channel & Vref from the previous call. Subsequent 
 * calls using the same settings will execute slightly faster (FI_FVR_DLY -
 * S_TM_ADC_ACQ) since settle time for Vref isn't necessary.
 * 
 * S_RT_ADC_MV_M provides the maximum range at the expense of accuracy because
 * the Vref value used (Vdd of 5V) does not evenly divide into the ADC's full-
 * scale value. The error is approximately 117uV/bit or full scale ~ 119.691mV. 
 * Depending on the scale used the total error could be significant. The 
 * constant S_RT_MV_M_COR compensates for this but does not fully eliminate the 
 * error. The value should be adjusted to obtain the most accurate values for 
 * the scale of the voltages being measured. The starting value is derived by 
 * dividing the integer resolution (5mV) by the 1 LSB error (0.005 - 0.00488281
 * =0.000117).
 * 
 * Returns:  none.  The following values are populated in object rv:
 *      vmv=composite value, resolution defined by S_RT_VMV_INT & S_RT_VMV_DEC
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=unmodified value read from ADC; not scaled
 *      mvr=mV per ADC bit set (unmodified)
 *
 * Parameters:  channel=analog channel.  The pin must be properly configured.
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts: rv.mvr must be set to 
 *                  S_RT_ADC_MV_L, S_RT_ADC_MV_H, or S_RT_ADC_MV_M.
 *
*/
void ReadVoltage(unsigned char channel, unsigned char scale, volts *rv)
{
    static union {
        unsigned char result[S_SZ_ADC_INT];
        unsigned int value;
    } adc;
    static unsigned char adcr = S_RT_ADC_MV_U;  //Current ADC range
    static unsigned int adc1v = S_QY_ADC_1V_M;  //ADC value for 1V @ max range

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Connect ADC to channel and acquire.  Assign Vref to proper FVR range or
     * Vdd, allow for settle time and setup conversion constants.  The ADC 
     * mV/bit resolution is stored in the user-supplied volts object for use by 
     * other conversion functions such as CalcAvg and should not be modified.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0 = (unsigned) (channel | S_EN_ADC);   //Connect to channel & enable ADC
    if (rv->mvr != adcr)
    {
        if (rv->mvr == S_RT_ADC_MV_L)           //Low range, high res
        {
            ADCON1 |= S_MK_ADC_FVR;             //Assign Vref
            FVRCONbits.ADFVR = CFG_FVR_ADC_2X;  //Set FVR
            adc1v = S_QY_ADC_1V_L;              //ADC value/Volt
        }
        else if (rv->mvr == S_RT_ADC_MV_H)      //High range, medium res
        {
            ADCON1 |= S_MK_ADC_FVR;             //Assign Vref
            FVRCONbits.ADFVR = CFG_FVR_ADC_4X;  //Set FVR
            adc1v = S_QY_ADC_1V_H;              //ADC value/Volt
        }
        else                                    //Max range, low res
        {
            ADCON1 &= S_MK_ADC_VDD;             //Assign Vref
            adc1v = S_QY_ADC_1V_M;              //ADC value/Volt
        }
        __delay_us(FI_FVR_DLY);                 //Refresh S&H cap + Vref settle
        adcr = rv->mvr;
    }
    else __delay_us(S_TM_ADC_ACQ);              //Refresh S&H cap

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Perform conversion, calculate, scale, and store result.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0bits.GO = TRUE;                       //Start conversion
    while (ADCON0bits.GO);                      //Wait for completion
    adc.result[POS_ADC_HB] = ADRESH;
    adc.result[POS_ADC_LB] = ADRESL;
    rv->adc = adc.value;                        //Save raw ADC value to user object
    if (adcr == S_RT_ADC_MV_M)                  //Apply correction for max range
        adc.value -= (adc.value / S_RT_MV_M_COR);
    adc.value *= scale;                         //Scale value
    rv->volt = (adc.value / adc1v);
    rv->mv = (((adc.value % adc1v) * rv->mvr) / S_RT_VMV_DEC);
    rv->vmv = ((rv->volt * S_RT_VMV_INT) + rv->mv);

    return;
} //End function ReadVoltage


/* Function CalcAvg.  Calculates the arithmetic mean of a sample set of 
 * previously acquired values.  No bounds checking is performed.
 * 
 * The value to be averaged is expected to represent the sum of raw ADC
 * conversions; they should not be scaled.  The average is calculated, scaled,
 * and the result stored in the object of type volts pointed to by rv.
 * Error correction (compensation) is also applied to the sum value if item mvr 
 * in the rv object is set to S_RT_ADC_MV_M. See function ReadVoltage for 
 * details.
 * 
 * Returns:  none.  The following values are populated in object rv.  All 
 * represent averaged values:
 *      vmv=composite value, resolution defined by S_RT_VMV_INT & S_RT_VMV_DEC
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=averaged & scaled ADC value
 *      mvr=mV per ADC bit set (unmodified)
 *
 * Parameters:  sum=raw summed ADC values
 *              count=number of summed values
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.  Object mvr in rv must be
 *              initialized to the proper ADC resolution.  See ReadVoltage for
 *              details.
 *
*/
void CalcAvg(unsigned int sum, unsigned char count, unsigned char scale, \
        volts *rv)
{
    unsigned int adc1v;                         //ADC value for 1V

    if (rv->mvr == S_RT_ADC_MV_L) adc1v = S_QY_ADC_1V_L;
    else if (rv->mvr == S_RT_ADC_MV_H) adc1v = S_QY_ADC_1V_H;
    else                                    //Max range
    {
        adc1v = S_QY_ADC_1V_M;
        sum /= count;                       //Apply correction to AVERAGE value
        rv->adc = ((sum - (sum / S_RT_MV_M_COR)) * scale);
    }

    if (rv->mvr != S_RT_ADC_MV_M) rv->adc = ((sum / count) * scale);
    rv->volt = (rv->adc / adc1v);
    rv->mv = (((rv->adc % adc1v) * rv->mvr) / S_RT_VMV_DEC);
    rv->vmv = ((rv->volt * S_RT_VMV_INT) + rv->mv);

    return;
} //End function CalcAvg


/* Function chgDutyCycle.  Moves the duty cycle from the start to final value 
 * incrementally for smooth operation & transient control.  The constant
 * S_TM_DCDELAY sets the delay, in uS, between DC increments.
 * 
 * Returns:     none
 *
 * Parameters:  start=16-bit starting DC value
 *              final=16-bit final DC value
 *
*/
void chgDutyCycle(uint16_t start, uint16_t final)
{
    while (TRUE)
    {
        PWM5DC = start;
        PWM5LDCONbits.LDA = TRUE;
        if (start == final) break;

        if (start < final) ++start; else --start;
        __delay_us(S_TM_DCDELAY);
    }

    return;
} //End function chgDutyCycle


/* Function writeFlashRow.  Writes a row of flash.  A 16-bit checksum for data
 * is first calculated and stored in the last word of the row.  Hence, the user
 * only has (FLASHROWSZ - 2) bytes available for use.
 * 
 * This function can also be used to restore default NVM settings by leveraging
 * the 'BOOT & NVM PROTECTION AND MANAGEMENT' documented in function
 * Initialize.  To do so this function should be called with the data parameter
 * set to NULL.  When so the checksum word is cleared so that upon the next
 * boot the NVM defaults will be used since NVM data is considered invalid.
 * 
 * Returns:  none
 *
 * Parameters:  addr=14-bit address of row in flash to write.  Must point to
 *                  start of array.
 *              data=pointer to array of data to write to flash -OR-
 *                   NULL to clear the existing checksum but leave other data
 *                   in the row intact.
 *              
 *
*/
void writeFlashRow(uint16_t addr, unsigned char *data)
{
    uint16_t chksum;
    unsigned char count;

    //Writing full row
    if (data)
    {
        count = 0;                              //Start at beginning of row
        chksum = nvmChksum(data);               //Calculate & store checksum
        *(data + FLASH_AR_CSUM_L) = (unsigned char) (chksum & S_MK_U16_L);
        *(data + FLASH_AR_CSUM_H) = \
                (unsigned char) ((chksum & S_MK_U16_H) >> S_SZ_BYTE);

        //Must erase row first for clean write
        eraseFlashRow(addr);
    }
    //Clearing checksum only
    else
    {
        count = FLASH_AR_CSUM_L;                //Start at checksum offset in row
        chksum = 0;                             //Set checksum to zero
        data = (unsigned char *) &chksum;       //Set data address to checksum
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Setup to write data to flash.  Interrupts must be disabled for entire
     * sequence.  Set LWLO to load data to write latches and then to flash on
     * last byte.
     * 
     * Load high address register here & only once.  Update low register for 
     * individual bytes to write within row.
     * 
     * Variable count represents array element to write.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    di();                               //Interrupts disabled for entire sequence.
    WREN = TRUE;                        //Enable write
    CFGS = FALSE;                       //Program memory
    FREE = FALSE;                       //Write data
    LWLO = TRUE;                        //Load data latches
    PMADRH = (unsigned char) (addr >> 8);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Important:  count must be initialized at beginning of function based on 
     * intent.  It controls number of bytes to write and address offset within
     * NVM row.  data must point to full row or checksum only.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    for (; (count < FLASHROWSZ); ++count, ++data)
    {
        PMADRL = (unsigned char) (addr + count);
        PMDATL = *data;
        unlockSeq();                    //Unlock for write
    }

    LWLO = FALSE;                                   //Write row
    unlockSeq();                                    //Write flash

    //All done:  disable writes & re-enable interrupts
    WREN = FALSE;
    ei();

    return;
} //End function writeFlashRow


/* Function eraseFlashRow.  Erases a row of program flash memory.  Interrupts
 * are disabled during the erase.
 * 
 * Returns:  none
 *
 * Parameters:  addr=14-bit pointer to starting address for row.
 *
*/
void eraseFlashRow(uint16_t addr)
{
    PMADRL = (unsigned char) addr;
    PMADRH = (unsigned char) (addr >> S_SZ_BYTE);

    //Erase row
    di();                               //Disable interrupts for duration of function
    CFGS = FALSE;                       //Program memory
    FREE = TRUE;                        //Erase row
    WREN = TRUE;                        //Enable erase
    unlockSeq();                        //Unlock for erase
    WREN = FALSE;                       //Disable erase
    ei();                               //Enable interrupts

    return;
} //End function eraseFlashRow


/* Function nvmChksum.  Calculates the checksum for a row of data.  The 'row'
 * is an unsigned char array of length FLASHROWSZ since this function is
 * intended to be used data storage in NVM (program flash).
 * 
 * This function is intended to be used in conjunction with writeFlashRow and
 * assumes the checksum resides in the last two bytes of the row.  Hence, when
 * calculating the checksum only (FLASHROWSZ - 2) bytes are examined.
 * 
 * The checksum is a simple sum of byte values in the row starting with offset
 * 0.
 * 
 * Returns:  16-bit checksum value
 *
 * Parameters:  data=pointer to data row
 *              
 *
*/
uint16_t nvmChksum(unsigned char *data)
{
    uint16_t sum;
    unsigned char count;

    for (sum = 0, count=0; (count < (FLASHROWSZ - FLASH_SZ_CSUM)); ++count)
        sum += *(data + count);

    return(sum);
} //End function nvmChksum


/*Function InterruptExec.  MPU interrupt handler.  Manages all enabled
 * interrupts which should be ordered according to priority high-low:
 *      + CM2 over-current counter
 *      + Timer6 (system timer)
 *      + USART Rx (service port RS232 communications)
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void __interrupt() InterruptExec(void)
{
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * TIMER1 OVERFLOW ISR HANDLER: sets over-current overflow bit flag to 
     * indicate that timer1 has wrapped to zero. Must be reset by firmware.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (TMR1IE & TMR1IF)
    {
        Opdata.ovc_of = 1;
        TMR1IF = FALSE;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * SYSTEM TIMER. Non-precision timed background service. ~10 mS frequency.
     * Maintains timing objects:
     * + Opdata.msc counts 10mS increments
     * + Opdata.sc counts elapsed seconds since boot / overflow
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (TMR6IE & TMR6IF)
    {
        if (++Opdata.msc == S_QY_TMR6_1S)
        {
            Opdata.msc = 0;
            ++Opdata.sc;
        }
        TMR6IF = FALSE;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * USART HANDLER:  service RS232 console.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    usartISR();

    return;
} //End function InterruptExec


/*Function Initialize.  called immediately after boot by function main.  
 * Opdataures system clock, validates PIC boot reason & sets status, validates
 * NVM & loads to RAM, configures pins & peripherals, and configures interrupts.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization
    Opdata.sts = STS_BOOT;              //Start with assumption of clean boot...

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * System clock & watchdog config
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OSCCON = CFG_OSC_C;                 //Configure system clock
    while (!(HFIOFL));                  //Wait for oscillator to stabilize
    WDTCON = CFG_WDT_CON;               //Watchdog


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * BOOT & NVM PROTECTION AND MANAGEMENT.
     *
     * Recovering from CPU/firmware failures (e.g. watchdog, stack OF/UF, etc.)
     * and NVM corruption is critical for safety and field reliability.  While
     * neither should happen unless there's a true failure, statistically it's 
     * possible when operating in environmental extremes or the more likely
     * existence of a firmware bug.
     * 
     * The goal is to maintain safe operation, keep the unit intact, and record
     * the failure state (in that order).
     * 
     * DATA MANAGEMENT
     * A single row (32 bytes) of NVM data is used for the storage of tunable
     * parameters (e.g. current limit, output voltage, etc.).  Object nvData
     * resides in flash (NVM) and is initialized with default values.  During
     * normal operation object ramData contains a copy of nvData and is used
     * to update nvData.  It must also be initialized with the same default
     * values as nvData.
     * 
     * DATA VALIDATION
     * NVM data is validated with a simple 16-bit checksum stored in the last
     * two bytes of nvData.
     * 
     * BOOT PROCESS
     * The CPU boot status is always checked first to ensure that any abnormal
     * condition is identified before being cleared by subsequent execution.  
     * The PCON & STATUS registers are examined for reboot due to watchdog, 
     * stack OF/UF, and possibly other conditions.  If found the status byte
     * (Opdata.sts) is updated and it's expected that the main RUN loop halts
     * normal operation.
     * 
     * The NVM checksum is verified before attempting to load.  The checksum
     * value is read directly from NVM and compared to the NVM's calculated
     * checksum.  If they match nvData is copied to ramData.
     * 
     * If there's a mismatch ramData retains it's default values and the status
     * code will be updated if not already set by the CPU checks.
     * 
     * The remainder of the initialization function continues unaltered and
     * when complete the main function enters the RUN loop.  The charger
     * will not start if any status code other than BOOT is set.  A code block 
     * in the main function loop is responsible for storing error codes, 
     * resulting from the boot process, or otherwise, in NVM.
     * 
     * Since the MCU's architecture requires the entire NVM row to be written
     * it's critical that ramData contains valid data.  This is why it's not
     * updated with nvData when checksum's don't match.  By writing the status
     * code it also ensures that if NVM was truly corrupt it will now contain
     * valid data (provided there's not an MCU flash failure) and there's a 
     * chance of the unit returning to operation when power cycled.
     * 
     * Compile note:  braces are intentionally used even though each side of
     * the evaluation contains only one statement.  During testing it was 
     * found that the compiler did not interpret the evaluation correctly 
     * without them.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (((PCON & S_LM_PCON_EX) != S_LM_PCON_BT) || \
            ((STATUS & S_LM_STAT_EX) != S_LM_STAT_BT)) Opdata.sts = STS_MCU;

    if (nvmChksum((unsigned char *) nvData) != (unsigned int) ((((unsigned int) \
            nvData[FLASH_AR_CSUM_H]) << S_SZ_BYTE) | ((unsigned int) \
            nvData[FLASH_AR_CSUM_L])))
    {
        if (Opdata.sts == STS_BOOT) Opdata.sts = STS_NVM;
    }
    else
    {
        memcpy((void *) ramData, (void *) nvData, FLASHROWSZ);
    }


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZE CONTROL DATA:
     * + Opdata.msc is cleared
     * + Opdata.sc is cleared
     * + Opdata.s_msc is cleared
     * + Opdata.s_sc is cleared
     * + Opdata.in.mvr set to maximum range
     * + Opdata.out.mvr set to maximum range
     * + Opdata.vdd.mvr set to low range
     * + RDword set to ramData array to address NVM data as words for 16-bit
     *   values
     * + Enable individual pin weak pull-ups
     * + Snapshot flag set FALSE (no system event captured)
     * + Over-current counter overflow flag is cleared
     * + Opdata_ss is cleared
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Opdata.msc = 0;
    Opdata.sc = 0;
    Opdata.s_msc = 0;
    Opdata.s_sc = 0;
    Opdata.in.mvr = S_RT_ADC_MV_M;
    Opdata.out.mvr = Opdata.in.mvr;
    Opdata.vdd.mvr = S_RT_ADC_MV_L;
    RDword = (uint16_t *) ramData;
    OPTION_REGbits.nWPUEN = FALSE;
    flag_ss = FALSE;
    Opdata.ovc_of = 0;
    memset((void *) &Opdata_ss, (int) 0x00, sizeof(sysdat));


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * SET PORT CONFIGURATION:
     * + Set digital latch initial states
     * + Pin in/out-put states
     * + Analog / digital
     * + Weak pull-ups
     * + Open drain (tri-state)
     * + Slew rate
     * + Logic input levels
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PORTA = S_IN_PORTA;
    PORTB = S_IN_PORTB;
    PORTC = S_IN_PORTC;
    //Port A
    TRISA = IO_PORTA;
    ANSELA = AD_PORTA;
    WPUA = WP_PORTA;
    ODCONA = OD_PORTA;
    SLRCONA = SR_PORTA;
    INLVLA = IL_PORTA;
    //Port B
    TRISB = IO_PORTB;
    ANSELB = AD_PORTB;
    WPUB = WP_PORTB;
    ODCONB = OD_PORTB;
    SLRCONB = SR_PORTB;
    INLVLB = IL_PORTB;
    //Port C
    TRISC = IO_PORTC;
    ANSELC = AD_PORTC;
    WPUC = WP_PORTC;
    ODCONC = OD_PORTC;
    SLRCONC = SR_PORTC;
    INLVLC = IL_PORTC;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set the status LED so there is some indication that the unit has power
     * if it hangs before getting to the run loop.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    setledFAULT();


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * MAP PERIPHERAL PINS:
     * + Gate drive (COG1-A)
     * + External slope compensation reset (COG1-B)
     * + C2 output (over-current) sourced to pin to drive timer1 as over-current
     *   counter
     * + Timer5 gate connected to gate drive pin for duty cycle timing
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PPS_P_COG1_G = PPS_M_COG1_G;            //Gate drive
    PPS_P_COG1_R = PPS_M_COG1_R;            //External slope compensation reset
    PPS_P_C2_OUT = PPS_M_C2_OUT;            //Over-current signal
    T5GPPS = PPS_M_TMR_5G;                  //Timer5 gate
    T1CKIPPS = PPS_M_TMR_1C;                //Timer1 external clock

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Initialize USART for command console.  This should be done at end of 
     * port / pin config since it sets its respective pins for the USART.
     * 
     * Default is console echo OFF.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    usartSetRx(ENABLE);
    usartSetEcho(FALSE);


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE FVR & DACs:
     * FVR=ADC & DAC Vref
     * 1=Vref for voltage control loop
     * 2=current limit
     * 3=Vdd measurement
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    FVRCON = CFG_FVR_C;
    __delay_us(FI_FVR_DLY);

    DAC1CON0 = CFG_DAC_1_C;
    DAC1REFH = ramData[FLASH_AR_VSET_H];
    DAC1REFL = ramData[FLASH_AR_VSET_L];
    DAC1LD = TRUE;                              //Load reference value

    DAC2CON0 = CFG_DAC_2_C;
    DAC2REFH = ramData[FLASH_AR_CLIM_H];
    DAC2REFL = ramData[FLASH_AR_CLIM_L];
    DAC2LD = TRUE;                              //Load reference value

    DAC3CON0 = CFG_DAC_3_C;
    DAC3REF = CFG_DAC_3_R;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE ADC:  enable and connect to first measurement source in RUN 
     * loop. It's expected that all further management of ADC is done thru 
     * function ReadVoltage.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0 = CFG_ADC_CON_0;
    ADCON1 = CFG_ADC_CON_1;
    ADCON2 = CFG_ADC_CON_2;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE OP-AMPS:
     * 1=voltage control loop error amp
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OPA1CON = CFG_OPA_1_C;
    OPA1ORS = CFG_OPA_1_O;
    OPA1NCHS = CFG_OPA_1_N;
    OPA1PCHS = CFG_OPA_1_P;
    OPA1EN = TRUE;                          //Enable


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE COMPARATORS:
     * 1=internal current loop modulator (PWM) control
     * 2=over-current sense
     * 3=external current loop modulator (PWM) control
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM1CON0 = CFG_CM_1_C0;
    CM1CON1 = CFG_CM_1_C1;
    CM1NSEL = CFG_CM_1_N;
    CM1PSEL = CFG_CM_1_P;

    CM2CON0 = CFG_CM_2_C0;
    CM2CON1 = CFG_CM_2_C1;
    CM2NSEL = CFG_CM_2_N;
    CM2PSEL = CFG_CM_2_P;

    CM3CON0 = CFG_CM_3_C0;
    CM3CON1 = CFG_CM_3_C1;
    CM3NSEL = CFG_CM_3_N;
    CM3PSEL = CFG_CM_3_P;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE COG:
     * 1=gate drive & external ramp reset.  Loop state & ramp set from NVM.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    COG1DBR = CFG_COG_1_DBR;
    COG1DBF = CFG_COG_1_DBF;
    COG1BLKR = CFG_COG_1_BLKR;
    COG1BLKF = CFG_COG_1_BLKF;
    COG1PHR = CFG_COG_1_PHR;
    COG1PHF = CFG_COG_1_PHF;
    COG1ASD1 = CFG_COG_1_ASD1;
    COG1ASD0 = CFG_COG_1_ASD0;
    COG1RIS0 = CFG_COG_1_RIS0;
    COG1RIS1 = CFG_COG_1_RIS1;
    
    if (ramData[FLASH_AR_LOOP])                 //Closed loop
    {
        setloopClosed();                        //Set external or internal ramp
    }
    else setloopOpen();                         //Open loop
    
    COG1FIS1 = CFG_COG_1_FIS1;
    COG1RSIM0 = CFG_COG_1_RSIM0;
    COG1RSIM1 = CFG_COG_1_RSIM1;
    COG1FSIM0 = CFG_COG_1_FSIM0;
    COG1FSIM1 = CFG_COG_1_FSIM1;
    COG1CON1 = CFG_COG_1_CON1;
    COG1CON0 = CFG_COG_1_CON0;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE CLCS:
     * n/a
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE TIMERS:
     * + Assign timers to PWMs
     * 1=C2 output to count over-current events
     * 5=COG1-A (duty cycle) timing
     * 6=ISR timing service
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CCPTMRS = CFG_CCP_TMRS;

    TMR1 = 0;
    T1CON = CFG_TMR_1_CON;
    T1GCON = CFG_TMR_1_GCON;

    T5CON = CFG_TMR_5_CON;
    T5GCON = CFG_TMR_5_GCON;

    T6CLKCON = CFG_TMR_6_CLK;
    T6HLT = CFG_TMR_6_HLT;
    T6RST = CFG_TMR_6_RST;
    T6PR = CFG_TMR_6_PR;
    T6CON = CFG_TMR_6_CON;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE PWMS:
     * 5=master timing for COG1 drive. Sets maximum & open-loop duty cycle, 
     * (stored in NVM).
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM5INTE = CFG_PWM_5_INTE;
    PWM5CLKCON = CFG_PWM_5_CLK;
    PWM5LDCON = CFG_PWM_5_LDCON;
    PWM5OFCON = CFG_PWM_5_OFCON;
    PWM5PH = CFG_PWM_5_PHASE;
    PWM5DC = (uint16_t) ramData[FLASH_AR_DC];
    PWM5PR = CFG_PWM_5_PERIOD;
    PWM5OF = CFG_PWM_5_OFFSET;
    PWM5TMR = CFG_PWM_5_TIMER;
    PWM5CON = CFG_PWM_5_CON;                //Start


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE PRGS:
     * 1=slope compensation.  The ramp slope is set by ramData (NVM).
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PRG1FTSS = CFG_PRG_1_FTSS;
    PRG1RTSS = CFG_PRG_1_RTSS;
    PRG1CON2 = ramData[FLASH_AR_RAMP];
    PRG1INS = CFG_PRG_1_INS;
    PRG1CON1 = CFG_PRG_1_CON1;
    PRG1CON0 = CFG_PRG_1_CON0;
    

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZATION COMPLETE:
     * + Enable peripheral & global interrupts
     * + Enable peripheral interrupts serviced by ISR
     * + Start system timer
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Enable interrupts
    TMR1IE = TRUE;                      //Over-current counter overflow
    TMR6IE = TRUE;
    T6ON = TRUE;                        //System timer

    return;
} //End function Initialize



//End of file