/*

DESCRIPTION:
    Main include file for sepicproto.c - prototype 250W coupled inductor SEPIC.

COPYRIGHT:      (c) 2020, Brian Cornell, all rights reserved.

 TARGET MCU:    PIC16F1769

HW REVISION:    200713

HW RELEASE:     prototype

VERSION:        0.1

MODIFICATION HISTORY:

    2 Aug 20, BWC, created release.


*/



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Version constants & definitions.  Build number is defined in constant 
 * VERBUILD and maintained by shell script build.sh and included in build.h.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define VERMAJOR    0
#define VERMINOR    1


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Necessary includes.  Build.h contains the build number constant which is
 * maintained by the makefile.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifndef UINT_MAX
#include <limits.h>
#endif
#include "build.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Conditional compilation.  Uncomment directive to enable / disable features.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                  Configure system clock & Watchdog Timer
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_OSC_C       0x78            //16MHz, FOSC config word determines clock
#define CFG_WDT_CON     0x10            //Watchdog interval=256mS, off.


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * System wide constants & definitions.  Convention:  S_<TYPE>_<INFO>
 * Types:
 *  + DC=Duty Cycle
 *  + DF=Default value for a configurable parameter
 *  + EN=ENable bit
 *  + RT=RaTio or scaling factor
 *  + LM=LiMiting operational value
 *  + TM=TiMing
 *  + MP=Peripheral port bit MaPping
 *  + CH=character
 *  + IN=INitializing value
 *  + VL=VoLtage
 *  + QY=quantity (count) of something
 *  + CD=command data
 *  + BL=boolean value
 *  + MK=bit MasK
 *  + SZ=size of an object or data type
 * 
 * Info is user defined and should be descriptive.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_TM_MOD_F          105             //PWM (modulator) frequency
#define S_DC_MIN            5               //Start & min. open loop DC
#define S_DC_MAX            91              //Maximum allowable duty cycle
#define S_DC_RUN            S_DC_MAX        //Default running duty cycle
#define S_RT_VDD            4               //Vdd measurement ratio
#define S_RT_IN             14              //Input volts measurement ratio
#define S_RT_OUT            14              //Output volts measurement ratio
#define S_TM_TMR6           10              //Timer6 period in mS
#define S_QY_MS_S           1000            //Number of mS in one second
#define S_QY_TMR6_1S        (S_QY_MS_S / S_TM_TMR6) //# of timer6 periods=1S
#define S_IN_SAMPLES        1               //Starting value for measured samples counter
#define S_LM_SAMPLES        5               //Number of measured samples to average
#define S_LM_VDD_NVM        3               //Vdd > to allow NVM writes
#define S_LM_CLLOW          1               //Lowest allowable current limit value
#define S_LM_CLHIG          1023            //Highest allowable current limit value
#define S_LM_VLOW           690             //Lowest allowable voltage set value
#define S_LM_VHIG           940             //Highest allowable voltage set value
#define S_LM_SLHIG          31              //Highest allowable slope set value
#define S_LM_RLOW           0               //Lowest allowable ramp set value
#define S_LM_RHIG           31              //Highest allowable ramp set value
#define S_SZ_BYTE           8               //Number of bits in a byte
#define S_MK_U16_L          0x00FF          //Mask to access low byte of 16-bit unsigned
#define S_MK_U16_H          0xFF00          //Mask to access high byte of 16-bit unsigned
#define S_RT_PERCENT        100             //Used as divisor to calculate % of a dividend value
#define S_TM_DCDELAY        500             //# uS to wait between each DC increment
#define S_TM_MONINTVL       100             //# SysTimer clock ticks for monitor updates (1S)
#define S_TM_RST            200             //# mS to wait before reseting PIC
#define S_TM_INT            300             //# mS for system to stabilize at power-up
#define S_LM_PCON_EX        0xFA            //PCON bits to examine at boot
#define S_LM_PCON_BT        0x18            //PCON expected value at boot
#define S_LM_STAT_EX        0x18            //STATUS bits to examine at boot
#define S_LM_STAT_BT        0x18            //STATUS expected value at boot
#define S_TM_RESTART        250             //# x 10 = mS to delay restart
#define S_LM_OVCC_CNT       100             //Max over-current count per loop
#define S_LM_OVCC_HYS       200             //Over-current hysteresis
#define S_CH_SPACE          ' '             //White space


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Analog channel mapping:  channel=port=pin.  The values assigned to these
 * constants should match those used by ADC.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define ANLCAN0             0x00            //AN0=RA0=19
#define ANLCAN1             0x04            //AN1=RA1=18
#define ANLCAN2             0x08            //AN2=RA2=17
#define ANLCAN3             0x0C            //AN3=RA4=3
#define ANLCAN4             0x10            //AN4=RC0=16
#define ANLCAN5             0x14            //AN5=RC1=15
#define ANLCAN6             0x18            //AN6=RC2=14
#define ANLCAN7             0x1C            //AN7=RC3=7
#define ANLCAN8             0x20            //AN8=RC6=8
#define ANLCAN9             0x24            //AN9=RC7=9
#define ANLCAN10            0x28            //AN10=RB4=13
#define ANLCAN11            0x2C            //AN11=RB5=12
#define ANLCANS6            0x38            //Switched AN6
#define ANLCANS7            0x3C            //Switched AN7


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * PPS input pin definitions.  Map for MPU and use to assign peripheral input
 * pins.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define PPSRA0              0x00            //RA0
#define PPSRA1              0x01            //RA1
#define PPSRA2              0x02            //RA2
#define PPSRA3              0x03            //RA3
#define PPSRA4              0x04            //RA4
#define PPSRA5              0x05            //RA5
#define PPSRA6              0x06            //RA6
#define PPSRA7              0x07            //RA7
#define PPSRB0              0x08            //RB0
#define PPSRB1              0x09            //RB1
#define PPSRB2              0x0A            //RB2
#define PPSRB3              0x0B            //RB3
#define PPSRB4              0x0C            //RB4
#define PPSRB5              0x0D            //RB5
#define PPSRB6              0x0E            //RB6
#define PPSRB7              0x0F            //RB7
#define PPSRC0              0x10            //RC0
#define PPSRC1              0x11            //RC1
#define PPSRC2              0x12            //RC2
#define PPSRC3              0x13            //RC3
#define PPSRC4              0x14            //RC4
#define PPSRC5              0x15            //RC5
#define PPSRC6              0x16            //RC6
#define PPSRC7              0x17            //RC7


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                           PPS & direct pin mapping
 * 
 * For PPS output pin assignments both aliases for pin registers and constants 
 * mapping what peripheral connects to the pin.
 *      Pin register alias convention:  PPS_P_XXX_YY where
 *          XXX=peripheral name
 *          YY=channel or other detail; can be omitted
 *      Peripheral mapping convention:  PPS_M_XXX_YY same as alias
 * 
 * Input pins only require mapping constant.
 * 
 * For pins accessed directly or assigned in a peripheral's register (like ADC)
 * use the following convention:  PIN_XT_YYYYYY where
 *          X=(I)nput or (O)utput
 *          T=(A)nalog or (D)igital
 *          YYYYYY=user assigned mnemonic describing function
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Charger (COG) gate drive pin mapping
#define PPS_P_COG1_G    RB7PPS          //COG1 A maps to gate driver, pin 10
#define PPS_P_COG1_R    RC5PPS          //COG1 B maps to external slope reset, pin 5
#define PPS_P_C2_OUT    RC7PPS          //Comparator 2 (C2) output maps to pin 9
#define PPS_M_COG1_G    0x04            //Output COG1 A signal on selected pin
#define PPS_M_COG1_R    0x05            //Output COG1 B signal on selected pin
#define PPS_M_C2_OUT    0x19            //Output C2 signal on selected pin
#define PPS_M_TMR_5G    PPSRB7          //Timer5 gate pin=RB7, for duty cycle timing
#define PPS_M_TMR_1C    PPSRC7          //Timer1 clock pin=RC7, for over-current event count

#define PPS_M_LAT       0x00            //Remove peripheral from pin & output latch value

//Analog pin mapping
#define PIN_IA_IVOLT    ANLCAN2         //ADC sense - input volts
#define PIN_IA_OVOLT    ANLCAN11        //ADC sense - output volts

//LED output pins for fault, run, and power levels
#define PIN_OD_FAULT    RA4             //Fault
#define PIN_OD_RUN      RA5             //Run


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Port/pin configuration: I/O, analog/digital, slew rate, etc.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Port I/O: 1=input, 0=output
#define IO_PORTA    0b00001101
#define IO_PORTB    0b01110000
#define IO_PORTC    0b00001010
//Port Analog/Digital: 1=analog, 0=digital
#define AD_PORTA    0b00000100
#define AD_PORTB    0b01110000
#define AD_PORTC    0b00001110
//Port weak pull-up: 1=enabled, 0=disabled
#define WP_PORTA    0b00001000
#define WP_PORTB    0b00000000
#define WP_PORTC    0b00000000
//Open drain: 1=open drain, 0=push-pull
#define OD_PORTA    0b00000000
#define OD_PORTB    0b00000000
#define OD_PORTC    0b00000000
//Slew rate: 1=limited, 0=maximum
#define SR_PORTA    0b00000000
#define SR_PORTB    0b00000000
#define SR_PORTC    0b00000000
//Input level: 1=ST, 0=TLL
#define IL_PORTA    0b00000000
#define IL_PORTB    0b00000000
#define IL_PORTC    0b00000000
//Initial latch states: all outputs logic low
#define S_IN_PORTA  0b00000000
#define S_IN_PORTB  0b00000000
#define S_IN_PORTC  0b00000000


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * SYSTEM MACROS.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define setledFAULT()       PIN_OD_FAULT = TRUE; PIN_OD_RUN = FALSE
#define setledRUN()         PIN_OD_FAULT = FALSE; PIN_OD_RUN = TRUE
#define cogStatus()         ((COG1ASD0bits.ASE) ? FALSE : TRUE)

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Used by several console functions to remove whitespace between command and
 * parameter.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define rmvSpaces(cs)       for (cs += CMDSIZSTR; (*cs == S_CH_SPACE); ++cs)

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Closes the control loop with the set slope compensation ramp (external /
 * internal).
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define setloopClosed() if (ramData[FLASH_AR_RAMP_E]) \
COG1FIS0 = CFG_COG_1_FIS0_C_E; else COG1FIS0 = CFG_COG_1_FIS0_C_I

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Complement to setloopClosed: opens the loop & duty cycle runs @ set value.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define setloopOpen() COG1FIS0 = CFG_COG_1_FIS0_O


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FVR, DAC, and function Initialize configuration constants
 * DAC1:  output voltage set
 * DAC2:  current limit
 * DAC3:  Vdd (measured by ADC)
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_FVR_C       0xCF            //Enabled, no temp sense, CDAFVRx4, ADFVRx4
#define CFG_FVR_ADC_4X  0x03            //ADC buffer set to 4x gain
#define CFG_FVR_ADC_2X  0x02            //ADC buffer set to 2x gain


//DAC1:  voltage is set from NVM; default defined here
#define CFG_DAC_1_C     0x80            //Enabled, right justified, Vdd, Vss
#define S_IN_VSET       818             //Loaded into DAC for output volts
#define S_IN_VSET_H     (S_IN_VSET >> S_SZ_BYTE)
#define S_IN_VSET_L     (S_IN_VSET & S_MK_U16_L)

//DAC2:  voltage is set from NVM; default defined here
#define CFG_DAC_2_C     0x80            //Enabled, right justified, Vdd, Vss
#define S_IN_CLIM       990             //Loaded into DAC for current limiting
#define S_IN_CLIM_H     (S_IN_CLIM >> S_SZ_BYTE)
#define S_IN_CLIM_L     (S_IN_CLIM & S_MK_U16_L)

//DAC3: Vdd/4
#define CFG_DAC_3_C     0x80            //Enabled, (+)=Vdd, (-)=Vss
#define CFG_DAC_3_R     0x08            //Vdd / 4

#define FI_FVR_DLY      30              //Time delay in uS for FVR buffer to stabilize


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *              ADC configuration, acquisition, and value storage
 * 
 * This section defines a uniform method for data acquisition and storage.  The
 * ADC is configured to operate thru function ReadVoltage and should not be 
 * accessed directly once function Initialize is called at boot.  Acquired 
 * values are stored in an object of type volts.  The integer and decimal 
 * portions of a voltage are stored in discrete variables.
 * 
 * Since the PIC family lacks floating point capability a 16-bit unsigned int
 * variable can represent the combined integer & decimal components in a format
 * referred to as 'VMV' (Volts-MilliVolts) format.  This allows evaluation as
 * an integer with the benefit of including the decimal portion for improved 
 * resolution.  This is done by multiplying the integer and dividing the decimal
 * values by some factors of 10, in effect shifting integer values left and
 * truncating decimal positions.
 * 
 * More places shifted offers higher resolution but with more limited range
 * (e.g. overflow).  The constants S_RT_VMV_INT & S_RT_VMV_DEC define the use of 
 * VMV values and must be assigned the paired values:
 * # Decimal Places         S_RT_VMV_INT        S_RT_VMV_DEC
 *          0               1                   1000
 *          1               10                  100
 *          2               100                 10
 *          3               1000                1
 * 
 * Prior to calling function ReadVoltage the variable mvr of the volts object
 * passed to the function must be initialized to S_RT_ADC_MV_M, S_RT_ADC_MV_H,
 *  or S_RT_ADC_MV_L.  This defines the Vref (5.0, 4.096, or 2.048) used and 
 * trades resolution for voltage range.
 * 
 * When S_RT_ADC_MV_M is used Vdd must be 5V and should be tightly regulated. A 
 * full scale error of 120mV (or ~ 117uV/bit) is introduced because 5V is not
 * evenly divisible by the ADC's bit resolution. This is in addition to the 
 * ADC's accuracy specification.
 * 
 * The constant S_RT_MV_M_COR is used by the functions ReadVoltage & CalcAvg to
 * compensate for this but does not fully eliminate the error. The value should 
 * be adjusted to obtain the most accurate values for the scale of the voltages 
 * being measured. The starting value is derived by dividing the integer 
 * resolution (5mV) by the 1 LSB error (0.005 - 0.00488281 = 0.000117).
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//ADC peripheral configuration
#define CFG_ADC_CON_0       (PIN_IA_IVOLT | S_EN_ADC);  //Default channel & Enabled
#define CFG_ADC_CON_1       0xD0            //Result right justified, Fosc/16, Vss, Vdd
#define CFG_ADC_CON_2       0x00            //No auto conversion

#define S_EN_ADC            0x01            //Enable bit position in ADCON0 register
#define S_MK_ADC_VDD        0xFC            //Mask to set Vref=Vdd
#define S_MK_ADC_FVR        0x03            //Mask to set Vref=FVR
#define S_LM_ADC_VAL        1023            //Maximum value 10-bit ADC can return
#define S_QY_MV_V           1000            //Number of mV in one volt
#define S_RT_MV_M_COR       44              //Correction value for maximum range
#define S_RT_ADC_MV_M       5               //mV / ADC bit for maximum range
#define S_RT_ADC_MV_H       4               //mV / ADC bit for high range
#define S_RT_ADC_MV_L       2               //mV / ADC bit for low range
#define S_RT_ADC_MV_U       0               //Undefined range
#define S_QY_ADC_1V_M       (S_QY_MV_V / S_RT_ADC_MV_M) //ADC value for 1V maximum
#define S_QY_ADC_1V_H       (S_QY_MV_V / S_RT_ADC_MV_H) //ADC value for 1V high
#define S_QY_ADC_1V_L       (S_QY_MV_V / S_RT_ADC_MV_L) //ADC value for 1V low
#define S_TM_ADC_ACQ        20              //# uS required for S&H to acquire
#define S_SZ_ADC_INT        sizeof(unsigned int)    //# bytes in ADC integer
#define POS_ADC_HB          1               //ADC result high byte
#define POS_ADC_LB          0               //ADC result low byte
#define S_RT_VMV_INT        10              //Sets integer position of VMV value
#define S_RT_VMV_DEC        100             //Sets decimal places of VMV value
#define S_RT_VMV_MUL        (S_RT_VMV_INT * 10) //Scales VMV * VMV result

//Internal channels (peripherals)
#define ADC_CH_FVR          0x7C            //FVR
#define ADC_CH_DAC_1        0x78            //DAC1
#define ADC_CH_TMP          0x74            //Temperature indicator
#define ADC_CH_DAC_2        0x70            //DAC2
#define ADC_CH_DAC_3        0x6C            //DAC3
#define ADC_CH_DAC_4        0x68            //DAC4


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * OPA1 configuration for outer voltage control loop
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_OPA_1_C     0x80            //Enabled, no override or tristate
#define CFG_OPA_1_O     0x00            //Override not used
#define CFG_OPA_1_N     0x00            //(-) connect to OPA1IN0- pin 13
#define CFG_OPA_1_P     0x02            //(+) connect to DAC1_out


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure comparator 1 for internal current loop modulator (PWM) control
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CM_1_C0     0x8E            //Enabled, ZLF, hysteresis, async output
#define CFG_CM_1_C1     0x00            //No interrupts
#define CFG_CM_1_N      0x01            //(-) connects to IN1-
#define CFG_CM_1_P      0x08            //(+) connects to PRG1_output

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure comparator 2 for over-current sense
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CM_2_C0     0x8E            //Enabled, ZLF, hysteresis, async output
#define CFG_CM_2_C1     0x00            //No interrupts
#define CFG_CM_2_N      0x01            //(-) connects to IN1-
#define CFG_CM_2_P      0x0B            //(+) connects to DAC2_output

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure comparator 3 for external current loop modulator (PWM) control
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CM_3_C0     0x8E            //Enabled, ZLF, hysteresis, async output
#define CFG_CM_3_C1     0x00            //No interrupts
#define CFG_CM_3_N      0x03            //(-) connects to IN3-
#define CFG_CM_3_P      0x01            //(+) connects to IN1+


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Allocate timers to PWM modules
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CCP_TMRS    0x00            //PWM3, 4, CCP1 & 2 use timer2

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer1 configuration: counts over-current events via C2 output
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_1_CON   0x85            //Pin=async clk source, enabled
#define CFG_TMR_1_GCON  0x00            //Counting gate disabled

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer5 configuration: times COG1-A on-time to calculate duty cycle.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_5_CON   0x41            //Fosc, 1:1 prescale, sync, on
#define CFG_TMR_5_GCON  0xD0            //Gate enable, 1 polarity, single pulse, gate pin

#define S_QY_TMR_5_MAX  S_TM_MOD_F      //Maximum count for gate drive period
#define S_QY_TMR_5_1US  (_XTAL_FREQ / 1000000)  //Count for 1uS
#define S_TM_TMR_5_RUN  (((S_TM_MOD_F / S_QY_TMR_5_1US)*2)+2)   //Max allowed time for DC capture

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer6 configuration: provides interrupt timing for background caretaker 
 * that performs system monitor & control.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_6_CLK   0x05            //MFINTOSC
#define CFG_TMR_6_CON   0x19            //Off, 1:2 prescale, 1:10 post scale
#define CFG_TMR_6_HLT   0xA0            //Prescale syncs Fosc/4, on bit syncs \
                                          to timer clock, free running
#define CFG_TMR_6_RST   0x00            //No external reset sources
#define CFG_TMR_6_PR    250             //Period value to obtain 10mS period


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure 16-bit PWM5:  master and sets max DC
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PWM_5_CON           0x80            //Enabled, active high standard
#define CFG_PWM_5_INTE          0x00            //No interrupts
#define CFG_PWM_5_CLK           0x01            //No prescale, clock=HFINTOSC
#define CFG_PWM_5_LDCON         0x80            //Load triggering is immediate
#define CFG_PWM_5_OFCON         0x00            //Independent run mode
#define CFG_PWM_5_PHASE         0x0000          //Phase count
#define CFG_PWM_5_PERIOD        S_TM_MOD_F      //Period (frequency)
#define CFG_PWM_5_OFFSET        0x0000          //Offset count
#define CFG_PWM_5_TIMER         0x0000          //Starting value for timer


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure Complementary Output Generator 1 (COG1).
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_COG_1_CON0          0xD4            //Enable, load, HFINTOSC, HB
#define CFG_COG_1_CON1          0xC0            //Delay chain DB, active high

//Rising events
#define CFG_COG_1_RIS0          0x00            //None
#define CFG_COG_1_RIS1          0x02            //PWM5
#define CFG_COG_1_RSIM0         0x00            //None
#define CFG_COG_1_RSIM1         0x02            //PWM5

//Falling events
#define CFG_COG_1_FIS0_C_I      0x06            //CM1, CM2 - internal ramp closed lopp
#define CFG_COG_1_FIS0_C_E      0x0C            //CM2, CM3 - external ramp closed lopp
#define CFG_COG_1_FIS0_O        0x04            //CM2 - open lopp
#define CFG_COG_1_FIS1          0x02            //PWM5
#define CFG_COG_1_FSIM0         0x04            //Hi>low for CM2, low level CM1, CM3
#define CFG_COG_1_FSIM1         0x02            //Hi>low for PWM5

//Auto shutdown
#define CFG_COG_1_ASD0          0xB8            //Shutdown, B=high, A=low
#define CFG_COG_1_ASD1          0x00            //No auto-shutdown sources

//Delays & blanking
#define CFG_COG_1_DBR           0               //Rising dead band
#define CFG_COG_1_DBF           0               //Falling dead band
#define CFG_COG_1_BLKR          0               //Rising edge blanking
#define CFG_COG_1_BLKF          0               //Falling edge blanking
#define CFG_COG_1_PHR           0               //Rising event phase delay
#define CFG_COG_1_PHF           0               //Falling event phase delay


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure PRG1 for slope compensation
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PRG_1_CON0          0xB0            //Enabled, edge, slope, off
#define CFG_PRG_1_CON1          0x02            //Fall low, rise high
#define CFG_PRG_1_INS           0x00            //OPA1OUT
#define CFG_PRG_1_CON2          0               //Slope=200mV/uS
#define CFG_PRG_1_RTSS          0x07            //PWM5
#define CFG_PRG_1_FTSS          0x07            //PWM5


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                          Print & Format Constants
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                          Object definitions
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Object to store a measured voltage with 3-digit mV resolution
typedef struct {
    unsigned int    volt;                   //Voltage
    unsigned int    mv;                     //Millivolts (3 decimals)
    unsigned int    vmv;                    //Composite volts & mv represented as integer
    unsigned int    adc;                    //Unmodified value read from ADC
    unsigned char   mvr;                    //mV/ADC bit: see ADC section for values
} volts;

//System operating data
//To calculate DC% from measured DC:  Config.odcp = ((Config.oldc * 100) / (CFG_PWM_5_PERIOD + 1));
typedef struct {
    uint8_t         sts;                    //System status code; see status codes
    uint8_t         ovc_of;                 //ISR timer1 overflow - timer1>value
    uint8_t         cogti;                  //Integer component of COG on-time
    uint8_t         cogtd;                  //Decimal component of COG on-time
    uint8_t         msc;                    //ISR mS timer
    uint8_t         s_msc;                  //Static mS timer
    uint16_t        sc;                     //ISR seconds timer
    uint16_t        s_sc;                   //Static seconds timer
    uint16_t        cogc;                   //Timer count of COG duty cycle
    uint16_t        ovc;                    //Over-current count from timer1
    volts           in;                     //Input voltage
    volts           out;                    //Output voltage
    volts           vdd;                    //Logic (Vdd) supply voltage)
} sysdat;

//Console command array
const char *CmdList[] = {
    "cc",                                   //Clear over-current count
    "cn",                                   //Clear NVM error
    "cp",                                   //Clear snapshot
    "gc",                                   //Get over-current count
    "gd",                                   //Get (dynamic) duty cycle %
    "ge",                                   //Get set output volts
    "gi",                                   //Get set over-current limit
    "gl",                                   //Get loop setting
    "gm",                                   //Get ramp type (E | I)
    "gn",                                   //Get NVM status code
    "go",                                   //Get output volts
    "gp",                                   //Get snapshot of dynamic parameters
    "gr",                                   //Get program release
    "gs",                                   //Get system status
    "gt",                                   //Get autostart setting
    "gu",                                   //Get internal ramp setting
    "gv",                                   //Get input volts
    "gx",                                   //Get max/set duty cycle %
    "rd",                                   //Restore NVM defaults
    "rt",                                   //Reset MCU
    "sa",                                   //Set autostart 1 | 0
    "sc",                                   //Set current limit #
    "se",                                   //Set echo 1 | 0
    "sl",                                   //Set loop 1 | 0
    "sm",                                   //Set ramp type (E | I)
    "so",                                   //Set charger on
    "su",                                   //Set internal ramp rate
    "sv",                                   //Set output volts #
    "sx"                                    //Set max / open-loop duty cycle %
};

#define CMDSIZSTR   2                       //Size of command string

//Array location of commands
#define NUMOFCMDS   sizeof(CmdList) / sizeof(const char *)
#define CMDCLRCNT   0                       //Clear over-current count
#define CMDCLRNVM   1                       //Clear NVM error
#define CMDCLRSNP   2                       //Clear snapshot
#define CMDGETCNT   3                       //Get over-current count
#define CMDGETDCY   4                       //Get (dynamic) duty cycle
#define CMDGETOVS   5                       //Get set output volts
#define CMDGETCUR   6                       //Get set over-current limit
#define CMDGETLPS   7                       //Get loop setting
#define CMDGETRPS   8                       //Get ramp type (E | I)
#define CMDGETNVM   9                       //Get NVM status code
#define CMDGETOUT   10                      //Get output volts
#define CMDGETSNP   11                      //Get snapshot of dynamic parameters
#define CMDGETREL   12                      //Get program release
#define CMDGETSTS   13                      //Get system status
#define CMDGETAUT   14                      //Get autostart setting
#define CMDGETIRS   15                      //Get internal ramp setting
#define CMDGETINV   16                      //Get input volts
#define CMDGETMDC   17                      //Get max/set duty cycle %
#define CMDRSTDFL   18                      //Restore NVM defaults
#define CMDRSTMCU   19                      //Reset MCU
#define CMDSETAUT   20                      //Set autostart 1 | 0
#define CMDSETCUR   21                      //Set current limit #
#define CMDSETECH   22                      //Set echo 1 | 0
#define CMDSETLOP   23                      //Set loop 1 | 0
#define CMDSETRPS   24                      //Set ramp type (E | I)
#define CMDSETCON   25                      //Set charger on
#define CMDSETIRS   26                      //Set internal ramp rate
#define CMDSETVOL   27                      //Set output volts #
#define CMDSETDCY   28                      //Set max / open-loop duty cycle %

//Message, format, and label strings
const char *msgAck = "Ack\n",               //Command completed
           *msgNck = "Nck\n",               //Command completed but unsuccessful
           *fmtGpib = "%u.%u: %u.%u\n",     //GPIB inquiry data response format
           *fmtPrg = "%u.%u b%u\n";         //Firmware version response format


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * STATUS CODES AND STATE MANAGEMENT.
 * 
 * The constants STS_XXXX represent the status codes used throughout the 
 * control program and recorded in NVM (diagnostics). The object StsList also 
 * provides brief text for status display via the service port.
 * 
 * STATE MANAGEMENT refers to identifying the operational state of the unit.
 *  + RUN: a non-error state that includes the status codes BOOT, STOP, and RUN.
 *  + FAULT: a non-recoverable error condition that requires intervention. This
 *    state can only be exited by power cycle or reboot.
 * 
 * The undefined status value (STS_UND) bifurcates the status codes between
 * RUN and FAULT states to allow for efficient determination of operating state.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define STS_BOOT        0               //Boot
#define STS_STOP        1               //Stop
#define STS_RUN         2               //Run
#define STS_UND         3               //Undefined: <=RUN, >=FAULT
#define STS_MCU         4               //MCU fault (stack, watchdog, etc.)
#define STS_NVM         5               //Invalid NVM

const char *StsList[] = {
    "BOOT",
    "STOP",
    "RUN",
    "UNDF",
    "MCU",
    "NVM"
};


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Flash program memory definitions.  For non-volatile storage of configurable
 * parameters.
 * 
 * Word values must be stored in little endian (L|H) format.
 * 
 * Important:  flash architecture is PIC family dependent and the memory map /
 * constants must be reviewed when changing MCUs.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define FLASHROWSZ          32              //Size of flash row: PIC specific
#define FLASH_SZ_CSUM   (FLASHROWSZ - 2)    //Store checksum in last word of row

//Flash memory unlock sequence macro
#define unlockSeq() PMCON2 = 0x55; PMCON2 = 0xAA; WR = TRUE; NOP(); NOP()

//Row address for data:  this value must match default assignment below
#define FLASH_AR_AUTO       0               //Autostart if TRUE
#define FLASH_AR_STS        1               //Last status code
#define FLASH_AR_DC         2               //Max (PWM5) or open loop duty cycle
#define FLASH_AR_LOOP       3               //True if closed
#define FLASH_AR_CLIM_L     4               //Current limit (DAC2) low byte
#define FLASH_AR_CLIM_H     5               //Current limit (DAC2) high byte
#define FLASH_AR_CLIM       (FLASH_AR_CLIM_L / sizeof(uint16_t))
#define FLASH_AR_VSET_L     6               //Output volts (DAC1) low byte
#define FLASH_AR_VSET_H     7               //Output volts (DAC1) high byte
#define FLASH_AR_VSET       (FLASH_AR_VSET_L / sizeof(uint16_t))
#define FLASH_AR_STATUS     8               //STATUS register
#define FLASH_AR_RAMP       9               //Internal ramp (slope comp) setting
#define FLASH_AR_PCON       10              //PCON register
#define FLASH_AR_RAMP_E     11              //True to use external (rising) ramp)

//Checksum stored in little endian format in last word of row
#define FLASH_AR_CSUM_L (FLASHROWSZ - 2)
#define FLASH_AR_CSUM_H (FLASHROWSZ - 1)
#define FLASH_AR_CSUM   (FLASH_AR_CSUM_L / sizeof(uint16_t))

//Default data to initialize row with.  Positions 30 & 31 reserved for checksum.
#define S_DF_FLASH_0        FALSE
#define S_DF_FLASH_1        STS_UND
#define S_DF_FLASH_2        S_DC_RUN
#define S_DF_FLASH_3        FALSE
#define S_DF_FLASH_4        S_IN_CLIM_L
#define S_DF_FLASH_5        S_IN_CLIM_H
#define S_DF_FLASH_6        S_IN_VSET_L
#define S_DF_FLASH_7        S_IN_VSET_H
#define S_DF_FLASH_8        0
#define S_DF_FLASH_9        CFG_PRG_1_CON2
#define S_DF_FLASH_10       0
#define S_DF_FLASH_11       FALSE
#define S_DF_FLASH_12       0
#define S_DF_FLASH_13       0
#define S_DF_FLASH_14       0
#define S_DF_FLASH_15       0
#define S_DF_FLASH_16       0
#define S_DF_FLASH_17       0
#define S_DF_FLASH_18       0
#define S_DF_FLASH_19       0
#define S_DF_FLASH_20       0
#define S_DF_FLASH_21       0
#define S_DF_FLASH_22       0
#define S_DF_FLASH_23       0
#define S_DF_FLASH_24       0
#define S_DF_FLASH_25       0
#define S_DF_FLASH_26       0
#define S_DF_FLASH_27       0
#define S_DF_FLASH_28       0
#define S_DF_FLASH_29       0


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function declarations.  See source file for explanation of each function.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void main(void);
uint8_t ChargerON(void);
void ChargerOFF(void);
__bit getSnapshot(uint8_t);
void ReadVoltage(unsigned char, unsigned char, volts *);
void CalcAvg(unsigned int, unsigned char, unsigned char, volts *);
void chgDutyCycle(uint16_t, uint16_t);
void eraseFlashRow(uint16_t addr);
void writeFlashRow(uint16_t, unsigned char *);
uint16_t nvmChksum(unsigned char *);
void __interrupt() InterruptExec(void);
void Initialize(void);


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Preprocessor checks.  Maintain at end of main include file.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/



//End of file